/*
** Module   : file_system_helper_linux.c
** Abstract : implementation for detection of simbolic links, non-FIFO pipes and special devices.
**            This is used by filelys.c
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130329 Created initial version to implement Linux specifics for file system
**                  operations.
*/
#include "filesys.h"

/**
* Utility method to check the type of the given file. Linux version.
* There are 4 main types a file can be: Directory , File , Device and Unknown.
* In addition to the 4 main types there are 3 additive stats (they can be 
* added to the main type. For example: FH -> File hidden, DL -> Directory link.
* In some cases the additive status can only be added to a specific main type.
* Ex: the Progress behaviour for hidden attribute -> only hidden file are 
* marked as hidden, hidden directories are only marked as directory.
*
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  jc 
*         reference to the Java class who called this function
*         (from static context)
* @param  fname 
*         The file name to be checked passed from FileChecker.checkFileType()
*
* @return A wrapped integer representing the status code of the file.
*/
JNIEXPORT jint JNICALL checkFileType(JNIEnv *env, jclass jc, jstring fname) 
{
   const char *file_path = (*env)->GetStringUTFChars(env, fname, NULL);
   if (file_path == NULL) 
   {
      return -1; /* OutOfMemoryError already thrown */
   }
   char *filename;
   
   int status= UNKNOWN;
   int f_d = 0;
   
   f_d = open(file_path, O_RDONLY); 
  
   //Check if open() was successful 
   if (f_d == -1 ) 
   { 
      return -1; 
   } 
 
   struct stat st;
   
   // Use fstat to detect files and directories or links that point to files
   // and directories as files or directories
   if(fstat(f_d, &st)) 
   {
      close(f_d); 
      return -1; 
   }
   
   // check if regular file
   if (S_ISREG(st.st_mode)) 
   {
      // File status is standalone status
      status = FILE;
      // if file check if hidden 
      
      //extract the filename from the path
      filename = (strrchr(file_path, '/'))+1;
      // in unix hidden files don't have a property that determines if
      // they are hidden instead a hidden file will start with '.'
      //check if first letter is '.'
      if (*filename == '.')
      {
         // hidden status is additive status
         status |= HIDDEN;
       }
   }
   
   // check if file is directory
   else if (S_ISDIR(st.st_mode)) 
   { 
      status = DIR;
   } 
 
   // Close the file 
   close(f_d); 
 
    struct stat buf;
   //get symbolic link status
   //we can't use fstat for detecting symbolic links
   
   int lnk_stat = lstat(file_path, &buf); 
   
   (*env)->ReleaseStringUTFChars(env, fname, file_path);
   //if link status is -1 return -1
   if (lnk_stat==-1) return -1;

   
   // check for special devices
   if (S_ISCHR(buf.st_mode) || S_ISBLK(buf.st_mode))
   { 
      // status code for devices is standalone
      status = DEVICE;
   }
   // check for symbolic links
   // NOTE: A symbolic link that points to a directory or file will not be
   // detected as file or directory by the logic present here so this status 
   //cannot be made inclusive with file or directory
   if( S_ISLNK(buf.st_mode))
   {
      //status code for symbolic link is additive
      status |= SYMLINK;
   }
   
   // check for fifo pipe
   if (S_ISFIFO(buf.st_mode))
   {
      // Pipe code status is additive
      status |= PIPE;
   }
   
   return status; 
}

