/*
** Module   : process_helper_linux.c
** Abstract : implementation for process creation/waiting logic in libp2j.so native library which
**            is called via JNI
**            This is used by process.c
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130329 Created initial version to implement Windows specifics for file system
**                  operations.
*/

#include "process.h" 
#include <sys/wait.h>
#include <term.h>

/**
 * init - Linux version of set up signal handling to ignore job control signals
 *        (reflecteds in the child process as well)
 *
 * @param   env   pointer to an object providing access to JNI environment
 *
 * @param   cls   reference to the Java class who called this function
 *                (from static context)
 *
 */ 
JNIEXPORT void JNICALL init(JNIEnv* env, jclass class)
{
   // ignore SIGINT, SIGTSTP, SIGTTIN, SIGTTOU signals
   // this is done once and early enough and these signals remain ignored
   struct sigaction sa;
   sa.sa_handler = SIG_IGN;
   sa.sa_flags = 0;
   sigemptyset(&sa.sa_mask);

   sigaction(SIGINT,  &sa, NULL); 
   sigaction(SIGTSTP, &sa, NULL);
   sigaction(SIGTTIN, &sa, NULL);
   sigaction(SIGTTOU, &sa, NULL);
}

/**
 * pseudoTerminalLaunch - Linux version of spawning a child process redirecting its standard
 *                        input, output and error to particular file descriptors
 *                        (FileDescriptor Java instances) 
 *      
 *
 * @param   env       pointer to a structure that provides access to JNI
 *                    environment
 *
 * @param   cls       reference to the Java class who called this function
 *                    (from static context)
 *
 * @param   cmdline   the command line used for launching the child process  
 *
 * @param   in        FileDescriptor Java object for stdin
 *
 * @param   out       FileDescriptor Java object for stdout
 *
 * @param   err       FileDescriptor Java object for stderr
 *
 *
 * @param   silent  boolean flag controlling interactive mode
 *
 * @return  > 1  PID of spawned child process if successful
 *          < 0  on error
 */ 

JNIEXPORT jint JNICALL
   pseudoTerminalLaunch(JNIEnv* env, jclass cls, jobjectArray cmdline, jobject in,
                        jobject out, jobject err, jboolean silent)
{
   jint pid;
   
   // make sure we have valid argument references
   if (cmdline == NULL)
   {
      return -1;
   }
   
   // make sure we have a tty as stdin
   if (!isatty(STDIN_FILENO))
   {
      return -2;
   }
   
   // access redirection flags and create pipes
   int pipe_i[2];                // pipe for stdin
   if (in != NULL)
   {
      if (pipe(pipe_i) < 0)
      {
         return -3;
      }
   }
   
   int pipe_o[2];                // pipe for stdout
   if (out != NULL)
   {
      if (pipe(pipe_o) < 0)
      {
         if (in != NULL)
         {
            close(pipe_i[0]);
            close(pipe_i[1]);
         }
         return -4;
      }
   }
   
   int pipe_e[2];                // pipe for stderr
   if (err != NULL)
   {
      if (pipe(pipe_e) < 0)
      {
         if (in != NULL)
         {
            close(pipe_i[0]);
            close(pipe_i[1]);
         }
         if (out != NULL)
         {
            close(pipe_o[0]);
            close(pipe_o[1]);
         }
         return -5;
      }
   }

   // due to a bug in the JVM 1.4.2 and earlier, we must override any
   // signal handler for SIGCHLD that may exist because otherwise the JVM
   // may reap our process instead of letting us do it (see Sun J2SE bugs 
   // #4945203, #4785154 and #4763362)
   signal(SIGCHLD, SIG_IGN);

   // create a child process
   pid = (jint) fork();
   
   if (pid < 0)
   {
      // error, nothing more to do
      return -6;
   }
   else if (pid == 0)
   {
      // child process

      if (!silent)
      {
         // we have to leave the screen clear for the program being launched
         system("tput clear");

         // in interactive mode, we redirect stderr to stdout
         dup2(STDOUT_FILENO, STDERR_FILENO);
      }

      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need write end for child's stdin in child
         close(pipe_i[1]);
         
         // redirect this pipe to stdin
         if (pipe_i[0] != STDIN_FILENO)
         {
            if (dup2(pipe_i[0], STDIN_FILENO) != STDIN_FILENO)
            {
               // handle duplication problem
               exit(-1);
            }
            
            close(pipe_i[0]);    // not required after duplication
         }
      }
      
      if (out != NULL)
      {
         // don't need read end for child's stdout in child
         close(pipe_o[0]);
         
         // redirect this pipe to stdout
         if (pipe_o[1] != STDOUT_FILENO)
         {
            if (dup2(pipe_o[1], STDOUT_FILENO) != STDOUT_FILENO)
            {
               // handle duplication problem
               exit(-1);
            }
            
            close(pipe_o[1]);    // not required after duplication
         }
      }
      
      if (err != NULL)
      {
         // don't need read end for child's stderr in child
         close(pipe_e[0]);
         
         // redirect this pipe to stderr
         if (pipe_e[1] != STDERR_FILENO)
         {
            if (dup2(pipe_e[1], STDERR_FILENO) != STDERR_FILENO)
            {
               // handle duplication problem
               exit(-1);
            }
            
            close(pipe_e[1]);    // not required after duplication
         }
      }

      // must build the parameter list
      int len = (int) (*env)->GetArrayLength(env, cmdline);
      int i;
      
      char** cmds = malloc(sizeof(char*) * (len + 1)); 
      char** current = cmds;
      
      // copy the strings into our own array
      for (i = 0; i < len; i++, current++)
      {
         jobject next = (*env)->GetObjectArrayElement(env, cmdline, i); 
         char*   tmp  = (char*) (*env)->GetStringUTFChars(env,
                                                          (jstring) next,
                                                          NULL);
         
         if (tmp == NULL)
         {
            // out of memory!
            exit(-1);
         }
         
         *current = strdup(tmp);
         
         (*env)->ReleaseStringUTFChars(env, (jstring) next, tmp);
      }
      
      // terminate the argv[] with a null pointer (current will be pointing
      // at cmds[len] after the loop ends)
      *current = NULL;

      // exec (never returns unless there is an error)
      int rc = execvp(*cmds, cmds);
      
      // if we are here, there was a problem!
      fprintf(stderr, "ERROR: execvp rc = %i; errno = %i\n", rc, errno);
      exit(-2);
   }
   else
   {
      // parent process
      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need read end for child's stdin in parent
         close(pipe_i[0]);
         
         // pass the write end back
         initializeFileDescriptor(env, in, pipe_i[1]); 
      }
      
      if (out != NULL)
      {
         // don't need write end for child's stdout in parent
         close(pipe_o[1]);
         
         // pass the read end back
         initializeFileDescriptor(env, out, pipe_o[0]); 
      }
      
      if (err != NULL)
      {
         // don't need write end for child's stderr in parent
         close(pipe_e[1]);
         
         // pass the read end back
         initializeFileDescriptor(env, err, pipe_e[0]); 
      }
   }
   
   return pid;
}

/**
 * pseudoTerminalWait - The Linux version of waits for a child process to terminate. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment
 *
 * @param   cls     reference to the Java class who called this function from
 *                  a static context
 *
 * @param   pid     PID of the process to be waited
 *
 * @param   silent  boolean flag to control interactive mode
 *
 */ 
JNIEXPORT void JNICALL pseudoTerminalWait(JNIEnv* env, jclass cls, int pid, jboolean silent)
{
   pid_t child = (pid_t) pid;
   int   status;

   waitpid(child, &status, 0);

   if (!silent)
   {
      // in interactive mode we're calling functions in NCURSES directly, 
      // as opposed to executing command line utilities for the same purpose
      // (NCURSES library is a requirement anyway so we want to have a more
      // efficient and more controlled use of that interface as well, while 
      // also avoiding the hard requirement that other utility programs be
      // installed in addition to P2J)

      // put "Press space bar to continue." message on the last line

      // move cursor to the beginning of the last terminal line 
      char *s;
      s = tparm(tigetstr("cup"), tigetnum("lines"), 0, 0, 0, 0, 0, 0, 0, 0);
      putp(s);

      // display prompt
      printf("Press space bar to continue.");

      // move cursor to the end of the last terminal line (as in Progress)
      s = tparm(tigetstr("cup"), tigetnum("lines"), tigetnum("cols"),
                0, 0, 0, 0, 0, 0, 0);
      putp(s);
   }
}


